#!/bin/bash
#
# Copyright (C) 2016 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

set -e

basedir="$(dirname "$(readlink -f $0)")"
arch=amd64
model_dir="$basedir/models"
model_name=pc-amd64
model_file="$model_dir/${model_name}.model"
channel=candidate
image_size=4G
convert_qcow2=0

usage() {
    cat <<EOF
Usage: $(basename $0) [OPTIONS...]
Create a custom core image with a preconfigured user.

Options
  -h,--help     This help
  -m,--model    Name of a model or path to the model assertion file. Default: $model_name
  -c,--channel  The snap channel to use. Default: $channel
  -s,--snap     Extra snaps to install.
  -S,--image-size
                The size of the generated disk image file. Default: $image_size
  -k,--key-file Path to ssh public key file to install.
  -Q,--convert-qcow2
                Convert the final raw image to qcow2 format.
  -d,--debug    Enable debug mode
EOF

    exit 1
}

TEMP=$(getopt -o hdm:c:s:S:k:Q --long help,debug,model:,channel:,snap:,key-file:,convert-qcow2 -- "$@")
eval set -- "$TEMP"

exec 2>&1

while true ; do
    case "$1" in
        -h|--help) 
            usage;;
        -d|--debug)
            set -x
            shift;;
        -m|--model)
            if [ -e "$2" ]; then
                model_file=$2
            else
                # Default to a model file in the list of known models
                model_file="$model_dir/$2.model"
                if [ ! -e "$model_file" ]; then
                    echo "ERROR: Model file does not exist for model $2"
                    exit 1
                fi
            fi
            shift 2;;
        -c|--channel)
            channel=$2
            shift 2;;
        -s|--snap)
            # Optionally provide the snap file you want pre-built into the image
            # as third parameter.
            snap=$2
            shift 2;;
        -S|--image-size)
            image_size=$2
            shift 2;;
        -k|--key-file)
            key_file=$2
            shift 2;;
        -Q|--convert-qcow2)
            convert_qcow2=1
            shift;;
        --) shift ; break ;;
        *) usage;;
    esac
done

if [ $(id -u) -ne 0 ] ; then
	echo "ERROR: This script must be executed as root"
	exit 1
fi

if [ -z "$key_file" ] ; then
    echo "WARNING: No ssh key inserted into image" 2>&1
    ssh_key_info="no ssh key"
else
    ssh_key_info="ssh key $key_file"
fi

model_name=$(basename "$model_file" .model)
arch=$(grep "^architecture:" "$model_file" | cut -d' ' -f2)

image_name=ubuntu-core-16-${model_name}.img
ubuntu_image_extra_args=

if [ ! -z "$snap" ] ; then
	ubuntu_image_extra_args="--extra-snaps $snap"
fi

echo "Creating a $image_size image '$image_name' from channel '$channel'"\
    "with model '$model_name' and $ssh_key_info"

ubuntu-image \
	--channel $channel \
	-o $image_name \
	--image-size $image_size \
	$ubuntu_image_extra_args \
	"$model_file"

kpartx -a $image_name
sleep 0.5 # TODO: fix?

loop_path=`findfs LABEL=writable`
tmp_mount=`mktemp -d`

mount $loop_path $tmp_mount

# Migrate all systemd units from core snap into the writable area. This
# would be normally done on firstboot by the initramfs but we can't rely
# on that because we  are adding another file in there and that will
# prevent the initramfs from transitioning any files.
core_snap=$(find $tmp_mount/system-data/var/lib/snapd/snaps -name "core_*.snap")
tmp_core=`mktemp -d`
mount $core_snap $tmp_core
mkdir -p $tmp_mount/system-data/etc/systemd
cp -rav $tmp_core/etc/systemd/* \
	$tmp_mount/system-data/etc/systemd/
sleep 1 # TODO: fix?
umount $tmp_core
rm -rf $tmp_core

# system-user assertion which gives us our qa user we use to
# log into the system
mkdir -p $tmp_mount/system-data/var/lib/snapd/seed/assertions
cp qa-user.assertion $tmp_mount/system-data/var/lib/snapd/seed/assertions

# Disable console-conf for the first boot
mkdir -p $tmp_mount/system-data/var/lib/console-conf/
touch $tmp_mount/system-data/var/lib/console-conf/complete

# Create systemd service which is running on firstboot and sets up
# various things for us.
mkdir -p $tmp_mount/system-data/etc/systemd/system
cat << EOF > $tmp_mount/system-data/etc/systemd/system/devmode-firstboot.service
[Unit]
Description=Run devmode firstboot setup
After=snapd.service snapd.socket

[Service]
Type=oneshot
ExecStart=/writable/system-data/var/lib/devmode-firstboot/run.sh
RemainAfterExit=yes
TimeoutSec=3min
EOF

mkdir -p $tmp_mount/system-data/etc/systemd/system/multi-user.target.wants
ln -sf /etc/systemd/system/devmode-firstboot.service \
	$tmp_mount/system-data/etc/systemd/system/multi-user.target.wants/devmode-firstboot.service

mkdir -p $tmp_mount/system-data/var/lib/devmode-firstboot
cat << EOF > $tmp_mount/system-data/var/lib/devmode-firstboot/00-snapd-config.yaml
network:
  version: 2
  ethernets:
    eth0:
      dhcp4: true
EOF

firstboot_file=$tmp_mount/system-data/var/lib/devmode-firstboot/run.sh
cat << 'EOF' > $firstboot_file
#!/bin/bash

set -e

# Don't start again if we're already done
if [ -e /writable/system-data/var/lib/devmode-firstboot/complete ] ; then
	exit 0
fi

echo "Start devmode-firstboot $(date -Iseconds --utc)"

if [ "$(snap managed)" = "true" ]; then
	echo "System already managed, exiting"
	exit 0
fi

# no changes at all
while ! snap changes ; do
	echo "No changes yet, waiting"
	sleep 1
done

while snap changes | grep -qE '(Do|Doing) .*Initialize system state' ;  do
	echo "Initialize system state is in progress, waiting"
	sleep 1
done

if [ -n "$(snap known system-user)" ]; then
	echo "Trying to create known user"
	snap create-user --known --sudoer
fi

cp /writable/system-data/var/lib/devmode-firstboot/00-snapd-config.yaml /writable/system-data/etc/netplan

# Enable console-conf again
rm /writable/system-data/var/lib/console-conf/complete

# Mark us done
touch /writable/system-data/var/lib/devmode-firstboot/complete

# Add public key
PUBKEY_REPLACE

# Reboot the system as its now prepared for the user
reboot
EOF
sed "s#PUBKEY_REPLACE#echo '$(cat $key_file)' > /home/qa/.ssh/authorized_keys#" -i  $firstboot_file

chmod +x $firstboot_file

cp $firstboot_file /tmp/max # TODO: fix

umount $tmp_mount
kpartx -d $image_name
rm -rf $tmp_mount

if [ "$convert_qcow2" -eq 1 ]; then
    qemu-img convert -O qcow2 "$image_name" "$image_name".qcow2
    mv "$image_name".qcow2 "$image_name"
fi
